﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace TCS.ChipToolkit.Cryptography.Emv
{
    public static class WeakPin
    {
        //Specification https://www.datagenetics.com/blog/september32012/index.html
        public static bool IsWeak(string pin)
        {
            if (ContainsDuplicates(pin))
            {
                return true;
            }

            if (IsSequence(pin))
            {
                return true;
            }

            if (IsPattern(pin))
            {
                return true;
            }

            if (IsPinPadPattern(pin))
            {
                return true;
            }

            if (IsTopUsedPin(pin))
            {
                return true;
            }

            if (IsLikelyBirthdate(pin))
            {
                return true;
            }

            return false;
        }

        private static bool ContainsDuplicates(string pin)
        {
            //Examples	1111|7777|9991|3000|1000

            //Only the first two digits have to be counted
            var digitsToCompare = new char[] { pin[0], pin[1] };

            foreach (var digitToCount in digitsToCompare)
            {
                var count = pin.Count(c => c == digitToCount);
                if (count > 2)
                {
                    return true;
                }
            }

            return false;
        }

        private static bool IsSequence(string pin)
        {
            // Examples: 0123|2345|7890
            // ascending - 1234
            // descending - 7654
            // incl. 0 counting as zero - 0123
            // even/odd sequence - 0246;2468;4680;1357;3579

            var digits = ConvertPinToIntArray(pin).ToArray();

            if (IsSequence(digits))
            {
                return true;
            }

            //Now check when 0 is counting as ten – 7890

            //Replace 0 with 10
            for (var i = 0; i < digits.Length; i++)
            {
                if (digits[i] == 0)
                {
                    digits[i] = 10;
                }
            }

            if (IsSequence(digits))
            {
                return true;
            }


            return false;
        }

        private static bool IsSequence(int[] digits)
        {
            //Calculate the differences between following digits

            var differences = new List<int>();
            for (var i = 0; i < digits.Length - 1; i++)
            {
                var difference = digits[i + 1] - digits[i];
                differences.Add(difference);
            }

            //Compare differences. They all must be the same.

            for (var i = 0; i < differences.Count - 1; i++)
            {
                if (!differences[i].Equals(differences[i + 1]))
                {
                    return false;
                }
            }

            return true;
        }

        private static bool IsPattern(string pin)
        {
            if (IsPatternAABB(pin))
            {
                return true;
            }

            if (IsPatternABAB(pin))
            {
                return true;
            }

            return false;
        }

        private static bool IsPatternABAB(string pin)
        {
            //any couples of 2 (ABAB) 1212|6969
            var parts = SplitInParts(pin, 2).ToArray();

            for (var i = 1; i < parts.Length; i++)
            {
                if (!parts[0].Equals(parts[i], StringComparison.Ordinal))
                {
                    return false;
                }
            }

            return true;
        }

        private static bool IsPatternAABB(string pin)
        {
            //any couples of 2 (AABB) 1122|6699

            if(pin[0] == pin[1] && pin[2] == pin[3])
            {
                return true;
            }

            return false;
        }

        private static bool IsPinPadPattern(string pin)
        {
            //1397 in any permutation -> 4 corners of the PIN pad

            if (pin.All(c => c is '1' or '3' or '9' or '7'))
            {
                return true;
            }

            //2580|0852 vertical number on PIN Pad in either direction

            if (pin.Equals("2580", StringComparison.Ordinal) || pin.Equals("0852", StringComparison.Ordinal))
            {
                return true;
            }

            return false;
        }

        private static bool IsLikelyBirthdate(string pin)
        {
            //likely year of birth of the drivers
            var upper = DateTime.Now.Year + 7; //+7 so the card expiry date also cant be used
            var lower = DateTime.Now.Year - 99;

            var iPin = Convert.ToInt32(pin);

            return iPin >= lower && iPin <= upper;
        }

        private static bool IsTopUsedPin(string pin)
        {
            //1004|2001 in top20 PIN list

            return pin.Equals("1004", StringComparison.Ordinal) || pin.Equals("2001", StringComparison.Ordinal);
        }

        private static IEnumerable<int> ConvertPinToIntArray(string pin)
        {
            foreach (var c in pin)
            {
                //Get char as int
                yield return c - '0';
            }
        }

        private static IEnumerable<string> SplitInParts(string s, int partLength)
        {
            for (var i = 0; i < s.Length; i += partLength)
            {
                yield return s.Substring(i, partLength);
            }
        }

    }
}
